// LZSS -  
//    .
// : O(d*p)
// : O(n*p*d)
//   :
//     n -  
//     p -  
//     d -  
//  d,p ,         ,
//    .        .
//    p  log(p),       
//    ,      .
// : 
//
// icq:133575351    email:lawnmower-man@mail.ru
//
// :
// DATA 1[7c] 			-  		c=1-128 
// REF  01[6c] [x8]		-  	=3- 66 ,  1-256 
// REF  00[6c] [x16]	- 			c=3- 66 ,  257-65792 

#include<stdio.h>
#include<malloc.h>
#include<memory.h>

#define OFFSET_MIN1 1				//!<   
#define OFFSET_MAX1 256				//!<   
#define OFFSET_MIN2 257				//!<   
#define OFFSET_MAX2 (257+65536)		//!<   
#define COUNT_MIN   3				//!<   
#define COUNT_MAX   (63+3)			//!<    
#define DATA_MIN	1				//!<   
#define DATA_MAX	128				//!<   
#define BUFFER_SIZE (65536*2)		//!<    (  ,  2^n)

class LZSS{
	int data_count;					//!<   
	int math_count;					//!<   
	int math_offset;				//!<   
	int index1,index2;				//!<    (,)
	char *buffer;					//!<  
	inline void assert(bool);
// 
	struct tree_item_st{
		int child;		//!<   
		int next;		//!<  
		int index;		//!<     
		int size;		//!<  
	}*tree;
	int tree_size;	//!<  ,   
	int tree_count;	//!<    
	int tree_root;	//!<  
	int tree_free;	//!<   
	void TreeInit();						//!<   
	int  TreeNewItem();						//!<   
	void TreeFreeItem(int);					//!<  
	void TreeRemove(int index);				//!<    
	int  TreeAdd(int index,int *offset);	//!<     
//    
	inline int GetSize();					//!<     
	inline int DiffIndex(int,int);			//!<      
	inline int IncIndex(int,int step=1);	//!<     
	inline int DecIndex(int,int step=1);	//!<     
//    
	void FlushData(FILE*);					//!<     
	void FlushMath(FILE*);					//!<     
public:
	LZSS();
	~LZSS();
	void Pack(FILE*,FILE*);					//!< 
	void Depack(FILE*,FILE*);				//!< 
	static void Pack(char*,char*);
	static void Depack(char*,char*);
};

void LZSS::assert(bool x)
{
//	if(!x) _asm int 3
}


LZSS::LZSS()
{
	tree_size=64;
	buffer=(char*)malloc(BUFFER_SIZE);
	tree=(tree_item_st*)malloc(tree_size*sizeof(tree_item_st));
}

LZSS::~LZSS()
{
	free(tree);
	free(buffer);
}

void LZSS::TreeInit()
{
	tree_root=tree_free=-1;
	tree_count=0;
}

//! \result   
int LZSS::TreeNewItem()
{
	int n;
	if(tree_free>=0) {
		n=tree_free;
		tree_free=tree[tree_free].next;
	} else {
		if(tree_count>=tree_size) 
			tree_size*=2,
			tree=(tree_item_st*)realloc(tree,tree_size*sizeof(tree_item_st));
		n=tree_count++;
	}
	tree[n].child=tree[n].next=-1;
	return n;
}

//! \param index   
void LZSS::TreeFreeItem(int index)
{
	tree[index].next=tree_free;
	tree_free=index;
}

//      ,    .
//       (  ,    )
//! \param index    
void LZSS::TreeRemove(int index)
{
	int i,n;
	int parent=-1;	//  
	int *ref=&tree_root;
	int root=tree_root;
	int count=COUNT_MAX;
	for(;;) {	//       
		for(;;) {	//   
			if(root<0) return;
			i=tree[root].index;
			if(buffer[i]==buffer[index]) break;
			ref=&tree[root].next; root=*ref;
		}
		for(n=0; n<tree[root].size; n++) {	//  
			if(buffer[i]!=buffer[index]) return;
			i=DecIndex(i);
			index=DecIndex(index);
		}
		count-=n;
		assert(count>=0);
		if(!count) break;
		parent=root; ref=&tree[root].child; root=*ref;
	}
	if(i!=index) return;	//       
	//   
	*ref=tree[root].next;
	TreeFreeItem(root);
	if(parent>=0) {	
		n=tree[parent].child;
		if(tree[n].next<0) {	//      ,  
			tree[parent].size+=tree[n].size;
			tree[parent].child=tree[n].child;
			TreeFreeItem(n);
		}
	}
}

//       .
// (       ,  )
//         .
//! \param index    
//! \retval offset     
//! \result  
int LZSS::TreeAdd(int index,int *offset)
{
	int n,i,nn,i0;
	int res=0;
	int limit=GetSize()+1; if(limit>COUNT_MAX) limit=COUNT_MAX;
	int count=COUNT_MAX;
	int parent=-1;
	int root=tree_root;	//   
	while(count) {
		for(;;) {	//   
			if(root<0) {	//   
				nn=TreeNewItem();
				if(parent<0) {		//   
					tree[nn].next=tree_root;
					tree_root=nn;
				}else{				//    
					tree[nn].next=tree[parent].child;
					tree[parent].child=nn;
				}
				tree[nn].index=index;
				tree[nn].size=count;
				goto exit;
			}
			i=tree[root].index;
			if(buffer[index]==buffer[i]) break;
			root=tree[root].next;
		}
		assert(i!=index);
		i0=index;
		for(n=0; n<tree[root].size; n++) {	//   
			if(buffer[index]!=buffer[i]) { //   
				nn=TreeNewItem();		//       (       )
				tree[nn].index=i;
				tree[nn].size=tree[root].size-n;
				tree[nn].child=tree[root].child;
				tree[root].child=nn;
				tree[root].size=n;
				break;
			}
			i=DecIndex(i);
			index=DecIndex(index);
		}
		count-=n;
		assert(count>=0);
		if(n==tree[root].size) tree[root].index=i0;	//   ,       
		parent=root,root=tree[root].child;	//    
		n=COUNT_MAX-count; 
		if(n>limit) n=limit;
		if(n<COUNT_MIN) n=0;
		if(n>res) res=n,*offset=DecIndex(index,i);
	}
exit:	
	return res;
}


void LZSS::FlushData(FILE* file)
{
	if(data_count<DATA_MIN) return;
	fputc((data_count-DATA_MIN)|0x80,file);
	for(int n=0; n<data_count; n++) {
		fputc(buffer[index1],file);
		index1=IncIndex(index1);
	}
	data_count=0;
}

void LZSS::FlushMath(FILE *file)
{
	if(math_count<COUNT_MIN) return;
	if(math_offset>OFFSET_MAX1) {
		math_offset-=OFFSET_MIN2;
		fputc((math_count-COUNT_MIN),file);
		fputc(math_offset&0xFF,file);
		fputc(math_offset>>8,file);
	}else{
		math_offset-=OFFSET_MIN1;
		fputc((math_count-COUNT_MIN)|0x40,file);
		fputc(math_offset,file);
	}
	index1=IncIndex(index1,math_count);
}


int LZSS::DiffIndex(int i1,int i2)
{
#if BUFFER_SIZE&(BUFFER_SIZE-1)	//     
	return (i1-i2+BUFFER_SIZE)%BUFFER_SIZE;
#else
	return (unsigned int)(i1-i2)&(BUFFER_SIZE-1);
#endif
}


int LZSS::GetSize()
{
	return DiffIndex(index2,index1);
}


int LZSS::IncIndex(int index,int step)
{
#if BUFFER_SIZE&(BUFFER_SIZE-1)	//     
	return (index+step)%BUFFER_SIZE;
#else
	return (unsigned int)(index+step)&(BUFFER_SIZE-1);
#endif
}


int LZSS::DecIndex(int index,int step)
{
#if BUFFER_SIZE&(BUFFER_SIZE-1)	//     
	return (index+BUFFER_SIZE-step)%BUFFER_SIZE;
#else
	return (unsigned int)(index-step)&(BUFFER_SIZE-1);
#endif
}


void LZSS::Pack(FILE *input,FILE *output)
{
	int print_count=0;
	int math_count,math_offset,sym,count;
	index1=index2=data_count=math_count=math_offset=count=0;
	memset(buffer,0,sizeof(buffer));
	TreeInit();
	for(;;) {
		this->math_count=math_count;
		this->math_offset=math_offset;
		sym=fgetc(input); if(sym==EOF) break;
		count++; if(count>print_count) printf("-%d   \r",count),print_count+=1243;
		TreeRemove(DecIndex(index2,OFFSET_MAX2));
		buffer[index2]=sym;
		math_count=TreeAdd(index2,&math_offset);
		index2=IncIndex(index2);
		if(this->math_count && math_count<=this->math_count) {
			data_count=GetSize()-this->math_count-1;
			FlushData(output);
			FlushMath(output);
			math_count=0;
		}
		data_count=GetSize();
		if(data_count==DATA_MAX) {
			FlushData(output);
			math_count=0;
		}
	}
	data_count=GetSize()-this->math_count;
	FlushData(output);
	FlushMath(output);
}


void LZSS::Depack(FILE *input,FILE *output)
{
	int sym,n;
	index1=index2=data_count=0;
	memset(buffer,0,sizeof(buffer));
	for(;;) {
		sym=fgetc(input); if(sym==EOF) break;
		if(data_count) {
			data_count--;
			buffer[index2]=sym;
			index2=IncIndex(index2);
			fputc(sym,output);
			continue;
		}
		if(sym&0x80) {data_count=(sym&0x7F)+DATA_MIN;continue;}
		math_count=(sym&0x3F)+COUNT_MIN; 
		math_offset=getc(input); if(math_offset==EOF) break;
		if(sym&0x40) math_offset++;
		else {
			sym=getc(input); if(sym==EOF) break;
			math_offset|=sym<<8;
			math_offset+=OFFSET_MIN2;
		}
		for(n=0; n<math_count; n++) {
			sym=buffer[DecIndex(index2,math_offset)]; buffer[index2]=sym;
			index2=IncIndex(index2);
			fputc(sym,output);
		}
	}
}


void LZSS::Pack(char *inname,char *outname)
{
	LZSS obj;
	printf("Pack: %s\n",inname);
	FILE *input=fopen(inname,"rb");
	FILE *output=fopen(outname,"wb");
	if(!input || !output) {printf("IO error\n");return;}
	obj.Pack(input,output);
	fclose(input);
	fclose(output);
}


void LZSS::Depack(char *inname,char *outname)
{
	LZSS obj;
	printf("Depack: %s\n",inname);
	FILE *input=fopen(inname,"rb");
	FILE *output=fopen(outname,"wb");
	if(!input || !output) {printf("IO error\n");return;}
	obj.Depack(input,output);
	fclose(input);
	fclose(output);
}


int main(int count,char **list)
{
	if(count!=4) {
		printf("%s -e infile outfile\n",list[0]);
		printf("%s -d infile outfile\n",list[0]);
		return 1;
	}
	if(list[1][0]!='-' || list[1][1]!='e' && list[1][1]!='d') {
		printf("Syntax error\n\a");
		return 2;
	}
	switch(list[1][1]) {
	case 'e': LZSS::Pack(list[2],list[3]); break;	// 
	case 'd': LZSS::Depack(list[2],list[3]); break;	// 
	}
	return 0;
}
